from functools import cached_property

from pydantic import BaseModel, computed_field

from ..hashing import HashDigest, get_hash
from ..utils import cat


class HashIdMixin(BaseModel):

    @computed_field(alias="_id", repr=True)
    @cached_property
    def id(self) -> HashDigest:
        return get_hash(cat(*self.hashable_fields()))

    def hashable_fields(self) -> list[str]:
        raise NotImplementedError



from pymongo.errors import PyMongoError


class RedBabyError(PyMongoError):
    """
    Generic error.
    """


class DocumentNotFound(RedBabyError):
    """
    Raised when no document is found.
    """


class ClientNotFoundError(Exception):
    def __init__(self, details: str = ""):
        self.details = details
        super().__init__(f"Redbaby Client not found. Details: {self.details}")


class ConnectionNotFoundError(Exception):
    def __init__(self, details: str = ""):
        self.details = details
        super().__init__(f"Redbaby Connection not found. Details: {self.details}")



from .core import BaseDocument
from .hashids import HashIdMixin
from .objectids import ObjectIdMixin
from .reading import ReadingMixin
from .timestamping import Timestamping

# from .versioning import Versioning
# from .softdeletion import SoftDeletion
#



from datetime import datetime, timezone

from pydantic import Field


class Timestamping:
    created_at: datetime = Field(default_factory=lambda: datetime.now(tz=timezone.utc))
    updated_at: datetime = Field(default_factory=lambda: datetime.now(tz=timezone.utc))



from redbaby.behaviors import ObjectIdMixin, ReadingMixin


class SomeDoc(ReadingMixin, ObjectIdMixin):
    attr: int

    @classmethod
    def collection_name(cls) -> str:
        return "some_doc"


def test_find():
    # setup
    SomeDoc.collection().insert_many([{"attr": 1}, {"attr": 2}, {"attr": 3}])
    # given
    filter = {"attr": 2}
    skip = 0
    limit = 1
    validate = True
    lazy = False
    sort = [("attr", 1)]

    # when
    result = SomeDoc.find(
        filter=filter,
        skip=skip,
        limit=limit,
        validate=validate,
        lazy=lazy,
        sort=sort,
    )

    # then
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], SomeDoc)
    assert result[0].attr == 2
    # cleanup
    SomeDoc.collection().delete_many({"attr": {"$in": [1, 2, 3]}})


def test_find_project():
    # setup
    SomeDoc.collection().insert_many([{"attr": 1}, {"attr": 2}, {"attr": 3}])
    # given
    filter = {"attr": 2}
    projection = {"attr": 0}
    limit = 1
    validate = False
    lazy = False
    # when
    result = SomeDoc.find(
        filter=filter,
        projection=projection,
        limit=limit,
        validate=validate,
        lazy=lazy,
    )

    # then
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], dict)
    assert "attr" not in result[0]
    assert "_id" in result[0]
    # cleanup
    SomeDoc.collection().delete_many({"attr": {"$in": [1, 2, 3]}})



from functools import partial
from typing import Any, Iterable, Literal, Optional, Self, Sequence, overload

from pymongo.cursor import CursorType

from .core import BaseDocument


class ReadingMixin(BaseDocument):
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        validate: Literal[False] = False,
        lazy: Literal[False] = False,
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> list[dict[str, Any]]: ...
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        validate: Literal[True] = False,
        lazy: Literal[True] = False,
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> Iterable[Self]: ...
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        validate: Literal[False] = False,
        lazy: Literal[True] = False,
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> Iterable[dict[str, Any]]: ...
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]] = None,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        validate: Literal[True] = False,
        lazy: Literal[False] = False,
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> list[Self]: ...

    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]] = None,
        skip: int = 0,
        limit: int = 0,
        validate: bool = False,
        lazy: bool = False,
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ):
        cursor_type = (
            CursorType.NON_TAILABLE if lazy or limit != 0 else CursorType.EXHAUST
        )
        col = cls.collection(alias=alias)
        cursor = col.find(
            filter=filter,
            projection=projection,
            skip=skip,
            limit=limit,
            sort=sort,
            cursor_type=cursor_type,
        )
        # TODO cache cursor when skip is not None
        validate_doc = cls.model_validate if validate else lambda x: x
        it = (validate_doc(item) for item in cursor)
        return it if lazy else list(it)



from typing import Optional, TypedDict

from pymongo import MongoClient
from pymongo.database import Database

from .errors import ClientNotFoundError, ConnectionNotFoundError


class MongoConnection(TypedDict):
    db_name: str
    uri: str


class DB:
    connections: dict[str, MongoConnection] = {}
    clients: dict[str, MongoClient] = {}

    @classmethod
    def get(
        cls,
        db_name: Optional[str] = None,
        suffix: Optional[str] = None,
        alias: str = "default",
    ) -> Database:
        if db_name is None:
            db_name = cls.get_conn(alias)["db_name"]
        if suffix:
            db_name = f"{db_name}-{suffix}"
        return cls.get_client(alias)[db_name]

    @classmethod
    def get_client(cls, alias: str = "default") -> MongoClient:
        try:
            return cls.clients[alias]
        except KeyError:
            raise ClientNotFoundError(alias)

    @classmethod
    def add_client(cls, alias: str = "default") -> MongoClient:
        cls.clients[alias] = MongoClient(host=cls.get_conn(alias)["uri"], fsync=True)

    @classmethod
    def get_conn(cls, alias: str = "default") -> MongoConnection:
        try:
            return cls.connections[alias]
        except KeyError:
            raise ConnectionNotFoundError(alias)

    @classmethod
    def add_conn(
        cls,
        db_name: str,
        uri: str,
        alias: str = "default",
        start_client: bool = True,
    ):
        cls.connections[alias] = MongoConnection(db_name=db_name, uri=uri)
        if start_client:
            cls.add_client(alias)


DB.add_conn(
    db_name="redbaby",
    uri="mongodb://localhost:27017",
    start_client=False,
)



from typing import Any, Iterable, Literal, Optional, Self, Sequence, overload

from .core import BaseDocument

class ReadingMixin(BaseDocument):
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]],
        skip: Optional[int],
        limit: Optional[int],
        validate: Literal[True],
        lazy: Literal[True],
        sort: Optional[Sequence[tuple[str, int]]],
        alias: str = "default",
    ) -> Iterable[Self]: ...
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]],
        skip: Optional[int],
        limit: Optional[int],
        validate: Literal[False],
        lazy: Literal[True],
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> Iterable[dict[str, Any]]: ...
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]],
        skip: Optional[int],
        limit: Optional[int],
        validate: Literal[False],
        lazy: Literal[False],
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> Sequence[dict[str, Any]]: ...
    @overload
    @classmethod
    def find(
        cls,
        filter: dict[str, Any],
        projection: Optional[dict[str, int | bool]],
        skip: Optional[int],
        limit: Optional[int],
        validate: Literal[True],
        lazy: Literal[False],
        sort: Optional[Sequence[tuple[str, int]]] = None,
        alias: str = "default",
    ) -> Sequence[Self]: ...



from redbaby.behaviors.core import BaseDocument
from redbaby.database import DB


class MyDoc(BaseDocument):
    attr: int

    @classmethod
    def collection_name(cls) -> str:
        return "my_doc"


def test_many_db_servers():
    # given
    DB.add_conn("db2", "mongodb://localhost:27017", alias="db2")
    DB.add_conn("db3", "mongodb://localhost:27017", alias="db3")
    # when
    MyDoc.collection(alias="db2").insert_one({"attr": 2})
    MyDoc.collection(alias="db3").insert_one({"attr": 3})
    # then
    db2_res = MyDoc.collection(alias="db2").find_one()
    db3_res = MyDoc.collection(alias="db3").find_one()
    assert db2_res is not None
    assert db2_res["attr"] == 2
    assert db3_res is not None
    assert db3_res["attr"] == 3
    # cleanup
    MyDoc.collection(alias="db2").delete_one({"attr": 2})
    MyDoc.collection(alias="db3").delete_one({"attr": 3})



from typing import Optional


def cat(*args: str, delimiter: str = "|") -> str:
    """

    >>> cat("foo", "bar", "baz")
    "foo|bar|baz"
    """
    return delimiter.join(args)


def get_root_dir(path: str) -> Optional[str]:
    """
    Get the root folder name from a path-like string.

    >>> get_root_dir("/foo/bar/baz")
    "foo"
    """
    without_surrounding_slashes = path.strip("/")
    breadcrumbs = without_surrounding_slashes.split("/")
    root_folder = breadcrumbs[0]
    if not root_folder:
        return None
    return root_folder


# from typing import Annotated, Any, Union

# from bson import ObjectId
# from pydantic import AfterValidator, PlainSerializer, WithJsonSchema


# def validate_object_id(v: Any) -> ObjectId:
#     if isinstance(v, ObjectId):
#         return v
#     if ObjectId.is_valid(v):
#         return ObjectId(v)
#     raise ValueError(f"Invalid ObjectId: {v!r}")


# PyObjectId = Annotated[
#     Union[str, ObjectId],
#     AfterValidator(validate_object_id),
#     PlainSerializer(str, return_type=str),
#     WithJsonSchema({"type": "string"}, mode="serialization"),
# ]


# class PyObjectId(BsonObjectId):
#     @classmethod
#     def __get_validators__(cls):
#         yield cls.validate

#     @classmethod
#     def validate(cls, v):
#         if not isinstance(v, (cls, BsonObjectId, str)) or not BsonObjectId.is_valid(v):
#             raise TypeError("Invalid ObjectId.")
#         return str(v)

#     @classmethod
#     def __get_pydantic_json_schema__(cls, field_schema: dict):
#         field_schema.update(
#             type="string",
#             examples=["5eb7cf5a86d9755df3a6c593", "5eb7cfb05e32e07750a1756a"],
#         )


def parse_sort(sort: str) -> list[tuple[str, int]]:
    """

    >>> parse_sort("-date,-count,+app")
    [("date", -1), ("count", -1), ("app", 1)]
    """
    return [
        (field_order[1:], -1 if field_order[0] == "-" else 1)
        for field_order in sort.split(",")
    ]



import shlex
import subprocess

import setuptools


def read_multiline_as_list(file_path: str) -> list[str]:
    with open(file_path, "r") as fh:
        contents = fh.read().split("\n")
        if contents[-1] == "":
            contents.pop()
        return contents


def get_version() -> str:
    raw_git_cmd = "git describe --tags"
    git_cmd = shlex.split(raw_git_cmd)
    fmt_cmd = shlex.split("cut -d '-' -f 1,2")
    git = subprocess.Popen(git_cmd, stdout=subprocess.PIPE)
    cut = subprocess.check_output(fmt_cmd, stdin=git.stdout)
    ret_code = git.wait()
    assert ret_code == 0, f"{raw_git_cmd!r} failed with exit code {ret_code}."
    return cut.decode().strip()


with open("README.md", "r") as fh:
    long_description = fh.read()
requirements = read_multiline_as_list("requirements.txt")


setuptools.setup(
    name="redbaby",
    version=get_version(),
    description="Mini reimplementation of REDB, baby redb, redbaby.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Nei Cardoso de Oliveira Neto",
    author_email="nei@teialabs.com",
    url="https://github.com/teialabs/redbaby",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    python_requires=">=3.11",
)
